// 
// Copyright (c) 2013-2016 Carsten Sonne Larsen <cs@innolan.dk>
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Globalization;
using System.IO;
using System.Net;
using Ntp.Common.Log;

namespace Ntp.Analyzer.Data.Import
{
    internal sealed class TimeServerLoader
    {
        internal TimeServerLoader(LogBase log)
        {
            this.log = log;
        }

        private const string NtpOrg = "http://support.ntp.org/bin/view/Servers/PublicTimeServer";
        private const string ServerTagStart = @"<div class=""twikiAfterText""></div><div class=""twikiForm"">";
        private const string ServerTagEnd = @"</div><!-- /twikiForm --></div><!-- /patternContent-->";
        private readonly LogBase log;

        public string ImportServer(int orgId)
        {
            if (orgId >= 10000)
            {
                log.WriteLine("Not contacting support.ntp.org for ID > 10000.", Severity.Debug);
                return null;
            }

            string html = GetOrgHtml(orgId);
            if (html == null)
            {
                log.WriteLine("No HTML received for TimeServer.", Severity.Debug);
                return null;
            }

            try
            {
                int serverTagPos = html.IndexOf(ServerTagStart, StringComparison.Ordinal);
                int tableStartPos = html.IndexOf("<table", serverTagPos, StringComparison.Ordinal);
                int tableEndPos = html.IndexOf(ServerTagEnd, serverTagPos, StringComparison.Ordinal);
                string table = html.Substring(tableStartPos, tableEndPos - tableStartPos);
                return table;
            }
            catch (Exception e)
            {
                log.WriteLine("Could not find TimeServer HTML.", Severity.Error);
                log.WriteLine(e);
                return null;
            }
        }

        private string GetOrgHtml(int orgId)
        {
            string url = NtpOrg + orgId.ToString(CultureInfo.InvariantCulture).PadLeft(6, '0');

            try
            {
                var client = new WebClient();
                client.Headers.Add(
                    "user-agent",
                    "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.2;)"
                    );

                Stream stream = client.OpenRead(url);
                log.WriteLine("Fetching data from support.ntp.org for server ID " + orgId, Severity.Info);

                if (stream != null)
                {
                    var reader = new StreamReader(stream);
                    string html = reader.ReadToEnd();
                    reader.Close();
                    return html.Replace("<nos>", string.Empty);
                }
            }
            catch (WebException e)
            {
                var response = e.Response as HttpWebResponse;
                if (response != null && response.StatusCode == HttpStatusCode.NotFound)
                {
                    log.WriteLine("TimeServer HTML page not found. ID: " + orgId, Severity.Warn);
                }
                else
                {
                    log.WriteLine("Error while contacting support.ntp.org with ID " + orgId, Severity.Error);
                }

                log.WriteLine(e);
            }
            catch (Exception e)
            {
                log.WriteLine("Error while contacting support.ntp.org with ID " + orgId, Severity.Error);
                log.WriteLine(e);
            }

            return null;
        }
    }
}